//   ( STL ,  )
// (c) Proteus     lawnmower-man@mail.ru

#include<stdio.h>
#include<stdlib.h>
#include<assert.h>

//!  
struct node_st{
	node_st(int);								
	int value;									//!<  
	node_st *next;
	node_st *child;
	static node_st *Merge(node_st*,node_st*);	//!<   
	void Clear();								//!<  
	void Dump(int depth);						//!<  
	int GetCount();								//!<    (  )
};


//!  (    )
class Vector{
	node_st **array; int size;
	void Resize(int size);
public:
	Vector();
	~Vector();
	int GetSize();
	node_st *&operator[] (int index);
	bool isEmpty();
};


//!  
class BiHeap{
	Vector vector;								//!<   
	void Insert(int h,node_st*);				//!<     
public:
	void Insert(int v);							//!<    
	bool isEmpty();								//!<   
	int  Pop();									//!<   
	void Dump();								//!<  
};


Vector::Vector()
{
	array=0; size=0;
}


Vector::~Vector()
{
	while(size--) if(array[size]) array[size]->Clear();
	free(array);
}

void Vector::Resize(int _size)
{
	int n=size;
	while(_size>=size) size+=1+size/4;
	if(n!=size) array=(node_st**)realloc(array,size*sizeof(node_st*));
	while(n<size) array[n++]=0;
}

node_st* &Vector::operator [](int index)
{
	Resize(index);
	return array[index];
}

bool Vector::isEmpty()
{
	for(int n=0; n<size; n++) if(array[n]) return false;
	return true;
}

int Vector::GetSize()
{
	return size;
}

int node_st::GetCount()
{
	int count=0;
	node_st *temp=child;
	while(temp) temp=temp->next,count++;
	return count;
}


//! \param v  
node_st::node_st(int v)
{
	value=v;
	next=child=0;
}

//!   ,      
node_st *node_st::Merge(node_st *node1,node_st *node2)
{
	node_st *node;
	assert(!node1->next && !node2->next);
	if(node1->value<node2->value) {
		node=node1;
		node2->next=node->child;
		node->child=node2;
	}else{
		node=node2;
		node1->next=node->child;
		node->child=node1;
	}
	return node;
}

void node_st::Clear()
{
	if(next) next->Clear();
	if(child) child->Clear();
	delete this;
}

//! \param depth  
void node_st::Dump(int depth)
{
	for(int n=0; n<depth; n++) putchar(' ');
	printf("[%d]\n",value);
	node_st *temp=child;
	while(temp) {
		temp->Dump(depth+2);
		temp=temp->next;
	}
}


//! \param v  
void BiHeap::Insert(int v)
{
	Insert(0,new node_st(v));
}

//! \param  
void BiHeap::Insert(int h,node_st *add)
{
	add->next=0;
	for(;;) {
		node_st *node=vector[h]; 
		if(!node) break;
		vector[h]=0;
		add=node_st::Merge(add,node);
		h++;
	}
	vector[h]=add;
}


int BiHeap::Pop()
{
	int n,index,min;
	node_st *node,*temp;
	index=-1;
	for(n=0; n<vector.GetSize(); n++) {
		node=vector[n];
		if(node && (index<0 || min>node->value)) min=node->value,index=n;
	}
	temp=vector[index];
	assert(temp->GetCount()==index);
	min=temp->value;
	node=temp->child;
	vector[index]=0; delete temp;
	while(node) {
		index--;
		assert(node->GetCount()==index);
		temp=node->next;
		Insert(index,node);
		node=temp;
	}
	return min;
}


bool BiHeap::isEmpty()
{
	return vector.isEmpty();
}

void BiHeap::Dump()
{
	printf("[dump]\n");
	for(int n=0; n<vector.GetSize(); n++) if(vector[n]) vector[n]->Dump(0);
}



#define SIZE 30

void main()
{
	BiHeap heap;
	for(int n=0; n<SIZE; n++) heap.Insert(rand()%30);
	heap.Dump();
	while(!heap.isEmpty()) printf("(%d)",heap.Pop());
	putchar('\n');
	printf("[press enter]\n");
	getchar();
}
